/*******************************************************
* TraceHandler.cpp
*
* Written by:     Ryan Fogarty
* Last Edit Date: 12-07-04
* Copyright:      2004
*
********************************************************/
#include "TraceHandler.h"

//////////////////////////////////////////////////////////
// The TraceHandlerManager Implementation
//////////////////////////////////////////////////////////
TraceHandlerManager::TraceHandlerManager()
{
   // Instantiate the Singleton<TraceHandler> implicitly
   // by calling the instance() function. Then start the
   // IceUtil::Thread right here!
   TraceHandler::instance().start();
}

TraceHandlerManager::~TraceHandlerManager()
{
   // DON'T MOVE THE ORDER OF ANY OF THESE STATEMENTS
   // OR SUFFER THE CONSEQUENCES!!!
   TraceHandler::instancePtr()->waitForStop();
   TraceHandler::destroy(); // Now kill it!
}
//////////////////////////////////////////////////////////
// End of the TraceHandlerManager Implementation
//////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////
// The TraceHandler Implementation
//////////////////////////////////////////////////////////
TraceHandler::TraceHandler(void)
{
   myThread = new TraceHandlerThread(this);
   myRedirectableReportStream = NULL;
   redirectToStdOut();
   std::cout << "TraceHandler has been created!\n";
}

TraceHandler::~TraceHandler(void)
{
   cleanup();
   std::cout << "TraceHandler has shutdown!\n";
}

void TraceHandler::start()
{
   if(myThread != NULL)
   {
      myThread->start();
   }
}

void TraceHandler::waitForStop()
{
   while(myThread->getThreadControl().isAlive())
   {
      // Extra brackets to ensure that theQueueLock goes out
      // of scope no matter the compiler.
      {
         IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
         isAlive = false;
         myQueueLockMonitor.notifyAll();
      }
   }
   // This below apparently does not work 100% of the time?
   // ?????????????????????????????????????????????????????
   // ?????????????????????????????????????????????????????
   // ?????????????????????????????????????????????????????
   //{
   //   std::cout << "Attempting to send shutdown signal!\n";
   //   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
   //   isAlive = false;
   //   myQueueLockMonitor.notifyAll();
   //   std::cout << "Sent shutdown notification!\n";
   //}
   //myThread->getThreadControl().join();
   // ?????????????????????????????????????????????????????
   // ?????????????????????????????????????????????????????
   // ?????????????????????????????????????????????????????
}

// file stream is not managed inside of class, but must be by application
void TraceHandler::redirectToFile(std::ofstream& aFileStream)
{
   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
   cleanup();
   if(aFileStream.is_open())
   {
      myRedirectableReportStream = new std::ostream(aFileStream.rdbuf());
   }
   else
   {
      std::cout << "ERROR: file stream is not open for tracing: " << std::endl;
      std::cout << "       redirecting to std::cout." << std::endl;
      redirectToStdOut();
   }
}

// New file is managed internally by the class
void TraceHandler::redirectToNewFile(std::string& aFileName)
{
   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
   cleanup();
   myOutFile.open(aFileName.c_str());
   if(myOutFile.is_open())
   {
      myRedirectableReportStream = new std::ostream(myOutFile.rdbuf());
   }
   else
   {
      std::cout << "ERROR: Cannot open file for tracing: " << aFileName << std::endl;
      std::cout << "       redirecting to std::cout." << std::endl;
      redirectToStdOut();
   }
}

void TraceHandler::redirectToStream(std::ostream& aStream)
{
   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
   cleanup();
   myRedirectableReportStream = new std::ostream(aStream.rdbuf());
}

void TraceHandler::redirectToStdOut()
{
   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
   cleanup();
   myRedirectableReportStream = new std::ostream(std::cout.rdbuf());
}

void TraceHandler::cleanup()
{
   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
   if(myRedirectableReportStream != NULL)
   {
      myRedirectableReportStream->flush();
      delete myRedirectableReportStream;
      myRedirectableReportStream = NULL;
      if(myOutFile.is_open())
      {
         myOutFile.close();
      }
      // else, do nothing
   }
}

void TraceHandler::writeOut(std::string& theMessage)
{
   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
   *myRedirectableReportStream << theMessage << std::endl;
}

void TraceHandler::put(std::string& aMessage)
{
   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
   myReportQueue.push(aMessage);
   if(myReportQueue.size() >= 1) myQueueLockMonitor.notify();
}

void TraceHandler::put(const char* aMessage)
{
   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor); //Lock the queue before writing to it!
   myReportQueue.push(aMessage);
   if(myReportQueue.size() >= 1)
   {
      // Sending a notify!
      myQueueLockMonitor.notify();
   }
}

// Virtual method as called by the Thread::start function
void TraceHandler::run()
{
   IceUtil::Monitor<IceUtil::RecMutex>::Lock theQueueLock(myQueueLockMonitor);
   isAlive = true;
   while(isAlive) // while theHandler is alive
   {
      while(isAlive && myReportQueue.empty())
      {
         myQueueLockMonitor.wait();
      }
      while(!myReportQueue.empty())
      {
         std::string theMessage = myReportQueue.front();
         myReportQueue.pop();
         writeOut(theMessage);
      }
   }
}
//////////////////////////////////////////////////////////
// End of the TraceHandler Implementation
//////////////////////////////////////////////////////////


#ifdef OTM_TRACE_TESTING
//////////////////////////////////////////////////////////
// The TraceHandlerUnitTester Implementation
//////////////////////////////////////////////////////////
void TraceHandlerUnitTester::run()
{
   TraceHandler::instancePtr()->redirectToNewFile(std::string("TestLog.txt"));
   OTM_TRACE("Write something to TestLog.txt file");
   OTM_TRACE("Write something else to TestLog.txt file");
   Sleep(10); // Give time for the TraceHandler thread to clear the queue
   TraceHandler::instancePtr()->redirectToStdOut();
   OTM_TRACE("Write something to std::out");
   OTM_TRACE("Write something else to std::out");
   Sleep(10); // Give time for the TraceHandler thread to clear the queue
   TraceHandler::instancePtr()->redirectToNewFile(std::string("TestLog2.txt"));
   OTM_TRACE("Write something to TestLog2.txt file");
   OTM_TRACE("Write something else to TestLog2.txt file");
   Sleep(10); // Give time for the TraceHandler thread to clear the queue
}
//////////////////////////////////////////////////////////
// End of the TraceHandlerUnitTester Implementation
//////////////////////////////////////////////////////////
#endif
