/*******************************************************
* TraceHandler.h
*
* Written by:     Ryan Fogarty
* Last Edit Date: 12-07-04
* Copyright:      2004
*
********************************************************/
#ifndef TRACE_HANDLER
#define TRACE_HANDLER

#include <queue>
#include <string>
#include "Singleton.h"
#include <iostream>
#include <fstream>
#include <IceUtil/Mutex.h>
#include <IceUtil/RecMutex.h>
#include <IceUtil/Monitor.h>
#include <IceUtil/Thread.h>

/** TraceHandler Class
 *    This class was created to provide a logging utility
 *    in which case Trace statements are directed to a user
 *    configurable output (either an existing ostream <of any variety>,
 *    an ofstream, a new managed ofstream, or simply std::cout).
 *
 * Usage:
 *    * Enable OTM_TRACE_HANDLING definition in compiler
 *    * Place in main(...):  TraceHandlerManager theManager;
 *    * Output messages:     OTM_TRACE(std::string("A simple string message..."));
 *                           OTM_TRACE("A simple const char* message");
 *
 * Unit-Tester Usage:
 *    * Enable OTM_TRACE_TESTING definition in compiler
 *    * Place in main(...): TRACE_UNIT_TEST();
 */

////////////////////////////////////////////////////
// Helper class to manage tracing thread
////////////////////////////////////////////////////
class TraceHandlerManager
{
public:
   TraceHandlerManager();
   ~TraceHandlerManager();
};

////////////////////////////////////////////////////
// Actual TraceHandler meat.
////////////////////////////////////////////////////
class TraceHandler : public Singleton<TraceHandler>
{
   friend Singleton<TraceHandler>;
   friend TraceHandlerManager;
public:
   // file stream is not managed inside of class, but must be by application
   void redirectToFile(std::ofstream& aFileStream);
   void redirectToNewFile(std::string& aFileName); // New file is managed internally by the class
   void redirectToStream(std::ostream& aStream);
   void redirectToStdOut();

   // Put onto the queue!
   void put(std::string& aMessage);
   void put(const char* aMessage);
protected:
   TraceHandler(void); // Use singleton to instantiate class!
   ~TraceHandler(void); // Use singleton to destroy object!
   // Implement Thread interface
   virtual void run();
private:
   // Override this method to disallow access!
   // Instead use TraceHandlerManager to manager object
   void start();
   void waitForStop();
   void cleanup();

   void writeOut(std::string& theMessage);
   std::queue<std::string> myReportQueue;
   std::ofstream myOutFile;
   std::ostream* myRedirectableReportStream;
   volatile 
      bool isAlive;
   IceUtil::Monitor<IceUtil::RecMutex> myQueueLockMonitor;

   ////////////////////////////////////////////////////
   // TraceHandlerThread
   //
   // Contribution by Mark Spruiell at ICE (THANKS!)
   //
   //   Allows TraceHandler class to manage Thread
   //   object via a Handle (Smart Pointer) and removes
   //   earlier multiple inheritance stipulation of 
   //   TraceHandler so that instead TraceHandler "has-a"
   //   Thread instead of "is-a" Thread.
   ////////////////////////////////////////////////////
   class TraceHandlerThread : public IceUtil::Thread
   {
   public:
      TraceHandlerThread(TraceHandler* aTH) : myTH(aTH) {}
      virtual void run() {myTH->run();}
   private:
      TraceHandler* myTH;
   };
   friend class TraceHandlerThread;
   IceUtil::ThreadPtr myThread;
};


//////////////////////////////////////////////////////
// Unit-Tester 
//////////////////////////////////////////////////////
#ifdef OTM_TRACE_TESTING
class TraceHandlerUnitTester
{
public:
   void run();
private:
   TraceHandlerManager theTraceHandlerManager;
};

// Create, run, and finally delete unit-tester
inline void TRACE_UNIT_TEST()
{
   TraceHandlerUnitTester *gTest = new TraceHandlerUnitTester();
   gTest->run();
   delete gTest;
}
// Ensure that the following *IS* defined for unit testing!
#define OTM_TRACE_HANDLING
#else
                          // Define to nothing!
#define TRACE_UNIT_TEST()
#endif // OTM_TRACE_TESTING

//////////////////////////////////////////////////////
// Definitions for simplified compile settable tracing
//////////////////////////////////////////////////////
#ifdef OTM_TRACE_HANDLING
                     // Use Tracing!
template<typename T>
inline void OTM_TRACE(T& aMessage)
{
   TraceHandler* pTH = TraceHandler::instancePtr(); 
   pTH != NULL ? 
      pTH->put(aMessage) : 
      std::cout << "WARNING TraceHandler not initialized: " << aMessage << std::endl;
}

#else
// Ignore Tracing!
                     // Null statement
#define OTM_TRACE(A) 
#endif // OTM_TRACE_HANDLING

#endif // #ifndef TRACE_HANDLER
